import unittest
from pprint import pprint

from pytolemaic.analysis_logic.model_analysis.sensitivity.sensitivity import \
    SensitivityAnalysis
from pytolemaic.analysis_logic.model_analysis.sensitivity.sensitivity_reports import SensitivityOfFeaturesReport, \
    SensitivityTypes


class TestSensitivity(unittest.TestCase):

    def test_sensitivity_meta(self):

        sensitivity = SensitivityAnalysis()
        sensitivities = {'a' + str(k): k for k in range(10)}

        mock = SensitivityOfFeaturesReport(method='mock', sensitivities=sensitivities,
                                           stats_report=sensitivity._sensitivity_stats_report(sensitivities))
        stats = mock.stats_report
        pprint(stats.to_dict())

        self.assertEqual(stats.n_features, 10)
        self.assertEqual(stats.n_zero, 1)
        self.assertEqual(stats.n_low, 1)

    def test_leakage(self):
        sensitivity = SensitivityAnalysis()
        self.assertEqual(sensitivity._leakage(n_features=10, n_very_low=0), 0)
        self.assertEqual(sensitivity._leakage(n_features=10, n_very_low=9), 1)
        self.assertGreaterEqual(sensitivity._leakage(n_features=10, n_very_low=8),
                                0.8)

        print([sensitivity._leakage(n_features=10, n_very_low=k) for k in
               range(10)])

    def test_overfit(self):
        sensitivity = SensitivityAnalysis()

        print([sensitivity._too_many_features(n_features=15, n_low=k+5, n_very_low=5+k//2, n_zero=5) for k in
               range(10)])

        self.assertEqual(
            sensitivity._too_many_features(n_features=10, n_low=0, n_very_low=0, n_zero=0), 0)
        self.assertEqual(
            sensitivity._too_many_features(n_features=10, n_low=5, n_very_low=0, n_zero=0), 0.5)

        self.assertGreater(
            sensitivity._too_many_features(n_features=10, n_low=9, n_very_low=9, n_zero=0), 0.9)


        self.assertGreaterEqual(
            sensitivity._too_many_features(n_features=10, n_low=9, n_very_low=9, n_zero=9), 0.0)

    def test_imputation(self):
        sensitivity = SensitivityAnalysis()

        shuffled = SensitivityOfFeaturesReport(method=SensitivityTypes.shuffled,
                                               sensitivities={'a': 0.3, 'b': 0.5, 'c': 0.2},
                                               stats_report=sensitivity._sensitivity_stats_report(
                                                   sensitivities={'a': 0.3, 'b': 0.5, 'c': 0.2}))
        missing = SensitivityOfFeaturesReport(method=SensitivityTypes.missing,
                                              sensitivities={'a': 0.3, 'b': 0.5, 'c': 0.2},
                                              stats_report=sensitivity._sensitivity_stats_report(
                                                  sensitivities={'a': 0.3, 'b': 0.5, 'c': 0.2}))

        self.assertEqual(sensitivity._imputation_score(
            shuffled=shuffled, missing=missing), 0)

        shuffled = SensitivityOfFeaturesReport(method=SensitivityTypes.shuffled, sensitivities={'a': 1, 'b': 0, 'c': 0},
                                               stats_report=sensitivity._sensitivity_stats_report(
                                                   sensitivities={'a': 1, 'b': 0, 'c': 0}))
        missing = SensitivityOfFeaturesReport(method=SensitivityTypes.missing, sensitivities={'a': 0, 'b': 1, 'c': 0},
                                              stats_report=sensitivity._sensitivity_stats_report(
                                                  sensitivities={'a': 0, 'b': 1, 'c': 0}))

        self.assertEqual(sensitivity._imputation_score(
            shuffled=shuffled, missing=missing), 1)



import os

HOME_DIR = os.path.join(os.path.dirname(__file__), '..')

from .pytrust import PyTrust
from .utils.constants import FeatureTypes, NUMERICAL, CATEGORICAL, REGRESSION, CLASSIFICATION
from .utils.dmd import DMD
from .utils.metrics import Metrics



import unittest
from pprint import pprint

import numpy

from pytolemaic import FeatureTypes
from pytolemaic.analysis_logic.dataset_analysis.covriance_shift import CovarianceShift
from pytolemaic.utils.dmd import DMD


class TestCovarianceShift(unittest.TestCase):

    def _gen_data(self, seed, offset=0.):
        rs = numpy.random.RandomState(seed)
        x = rs.randn(1000, 10)

        x[:, 0:5] = numpy.round(x[:, 0:5], 0)
        x[0, 0] = 10
        x[0, 9] = 10
        x[1, 9] = -4

        x[:,7] += offset
        y = numpy.copy(x[:, 0])

        x[10, :] = numpy.nan
        x[:, 1] = numpy.nan
        x[:700, 2] = numpy.nan

        return DMD(x=x, y=y, columns_meta={
          DMD.FEATURE_TYPES: 5 * [FeatureTypes.categorical] + 5 * [FeatureTypes.numerical]})

    def setUp(self) -> None:
        self.train = self._gen_data(0)
        self.same_dist_test = self._gen_data(1)
        self.mid_dist_test = self._gen_data(1, offset=1)
        self.dif_dist_test = self._gen_data(1, offset=3)

    def test_same_dist(self):
        covariance = CovarianceShift()
        covariance.calc_covariance_shift(dmd_train=self.train, dmd_test=self.same_dist_test)

        print(covariance.covariance_shift)
        assert covariance.covariance_shift < 0.03

        pprint(covariance.covariance_shift_report().to_dict())
        pprint(covariance.covariance_shift_report().insights())
        pprint(covariance.covariance_shift_report().plot())

    def test_mid_dist(self):
        covariance = CovarianceShift()
        covariance.calc_covariance_shift(dmd_train=self.train, dmd_test=self.mid_dist_test)

        print(covariance.covariance_shift)
        assert covariance.covariance_shift > 0.3
        assert covariance.covariance_shift < 0.6

        pprint(covariance.covariance_shift_report().to_dict())
        pprint(covariance.covariance_shift_report().insights())
        pprint(covariance.covariance_shift_report().plot())

    def test_high_dist(self):

        covariance = CovarianceShift()
        covariance.calc_covariance_shift(dmd_train=self.train, dmd_test=self.dif_dist_test)

        print(covariance.covariance_shift)

        assert covariance.covariance_shift > 0.75

        pprint(covariance.covariance_shift_report().to_dict())
        pprint(covariance.covariance_shift_report().insights())
        pprint(covariance.covariance_shift_report().plot())



if __name__ == '__main__':
    a=TestCovarianceShift()
    a.setUp()
    a.test_high_dist()
    from matplotlib import pyplot as plt
    plt.show()


import logging
import multiprocessing

import numpy
from sklearn.ensemble import RandomForestClassifier

from pytolemaic.utils.metrics import Metrics
from pytolemaic.utils.dmd import DMD
from pytolemaic.utils.general import GeneralUtils

from pytolemaic.analysis_logic.dataset_analysis.covariance_shift_report import CovarianceShiftReport



class CovarianceShiftCalculator():

    @classmethod
    def prepare_dataset_for_score_quality(cls, dmd_train: DMD, dmd_test: DMD):
        '''

        :param dmd_train: train set
        :param dmd_test: test set
        :return: dataset with target of test/train
        '''

        dmd = DMD.concat([dmd_train, dmd_test])
        new_label = [0] * dmd_train.n_samples + [1] * dmd_test.n_samples
        dmd.set_target(new_label)

        train, test = dmd.split(ratio=dmd_test.n_samples / (dmd_train.n_samples + dmd_test.n_samples))
        return train, test

    @classmethod
    def prepare_estimator(cls, train:DMD):
        classifier= GeneralUtils.simple_imputation_pipeline(
            estimator=RandomForestClassifier(random_state=0, n_estimators=30, n_jobs=multiprocessing.cpu_count() - 1))
        classifier.fit(train.values, train.target.ravel())
        return classifier

    @classmethod
    def calc_separation_quality(cls, classifier, test: DMD):
        yp = classifier.predict_proba(test.values)

        auc = Metrics.auc.function(y_true=test.target, y_pred=yp)
        auc = numpy.clip(auc, 0.5, 1)  # auc<0.5 --> 0.5

        #  High auc --> separable --> low quality.
        #  [0 to 1] --> [0.5 to 1] --> 1-(2*[0.5 to 1]-1) --> [1 to 0]
        separation_quality = numpy.round(1 - (2 * auc - 1), 5)
        return separation_quality

    @classmethod
    def calc_convriance_shift_auc(cls, classifier, test: DMD):
        return 1-cls.calc_separation_quality(classifier, test)

    @classmethod
    def calc_covariance_shift(cls, dmd_train: DMD, dmd_test: DMD):
        '''
        :param dmd_train: train set
        :param dmd_test: test set
        :return: estimation of score quality based on similarity between train and test sets
        '''

        train, test = cls.prepare_dataset_for_score_quality(dmd_train=dmd_train, dmd_test=dmd_test)
        classifier = cls.prepare_estimator(train=train)
        return cls.calc_convriance_shift_auc(classifier=classifier, test=test)

class CovarianceShift():
    def __init__(self):
        self._separation_quality = None
        self._cov_train = None
        self._cov_test = None
        self._classifier = None
        self._sensitivity = None

        self._dmd_train = None
        self._dmd_test = None

    def calc_covariance_shift(self, dmd_train: DMD, dmd_test: DMD):
        # save data for later report
        self._dmd_train = dmd_train
        self._dmd_test = dmd_test

        # split data to new train / test setes
        self._cov_train, self._cov_test = CovarianceShiftCalculator.prepare_dataset_for_score_quality(dmd_train=dmd_train, dmd_test=dmd_test)
        self._classifier = CovarianceShiftCalculator.prepare_estimator(train=self._cov_train)
        self._covariance_shift = CovarianceShiftCalculator.calc_convriance_shift_auc(classifier=self._classifier, test=self._cov_test)

    def covariance_shift_report(self):
        medium_lvl = 0.7
        high_lvl = 0.95
        if self.covariance_shift > medium_lvl:
            sensitivity_report=self.calc_sensitivity_report()
        else:
            sensitivity_report=None

        return CovarianceShiftReport(covariance_shift=self.covariance_shift,
                                     sensitivity_report=sensitivity_report,
                                     medium_lvl=medium_lvl, high_lvl=high_lvl,
                                     train=self._dmd_train,
                                     test=self._dmd_test)

    def calc_sensitivity_report(self):
        try:
            from pytolemaic.analysis_logic.model_analysis.sensitivity.sensitivity import SensitivityAnalysis
        except:
            logging.exception("Failed to import SensitivityAnalysis")
            return None

        self._sensitivity = SensitivityAnalysis()
        sensitivity_report = self._sensitivity.sensitivity_analysis(model=self.classifier, dmd_train=self.train, dmd_test=self.test,
                                               metric=Metrics.auc.name)
        return sensitivity_report

    @property
    def sensitivity(self):
        return self._sensitivity

    @property
    def separation_quality(self):
        return 1-self._covariance_shift

    @property
    def covariance_shift(self):
        return self._covariance_shift

    @property
    def train(self):
        return self._cov_train

    @property
    def test(self):
        return self._cov_test

    @property
    def classifier(self):
        return self._classifier



import unittest

from pytolemaic.analysis_logic.model_analysis.scoring.scoring_report import ConfusionMatrixReport, ScatterReport, \
    ScoringMetricReport, ScoringFullReport


class TestScoringReport(unittest.TestCase):

    def equal_to_dict_keys(self, rep):
        d1 = rep.to_dict()
        d2 = rep.to_dict_meaning()
        return d1.keys() == d2.keys()

    def test_to_dict_meaning(self):
        rep1 = ConfusionMatrixReport(y_true=[1,2,3], y_pred=[1,2,3])
        self.assertTrue(self.equal_to_dict_keys(rep1))

        rep2 = ScatterReport(y_true=[1,2,3], y_pred=[1,2,3])
        self.assertTrue(self.equal_to_dict_keys(rep2))

        rep3 = ScoringMetricReport(metric='mae', value=0.5, ci_low=0.25, ci_high=0.75)
        self.assertTrue(self.equal_to_dict_keys(rep3))

        rep4 = ScoringFullReport(target_metric='mae', metric_reports=[rep3],
                                 confusion_matrix=rep1, scatter=rep2)
        self.assertTrue(self.equal_to_dict_keys(rep4))




import multiprocessing
import unittest
from pprint import pprint

import numpy
import sklearn
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

from pytolemaic.prediction_uncertainty.uncertainty_model import \
    UncertaintyModelClassifier, UncertaintyModelRegressor
from pytolemaic.utils.dmd import DMD
from pytolemaic.utils.general import GeneralUtils


class TestPredictionsUncertainty(unittest.TestCase):

    def get_data(self, is_classification, seed=0):
        rs = numpy.random.RandomState(seed)
        x = rs.rand(10000, 4)
        x[:, 1] = 0
        # 1st is double importance, 2nd has no importance
        y = numpy.sum(x, axis=1) + 2 * x[:, 0]
        y[x[:, 0] < 0.2] = x[x[:, 0] < 0.2, 2]
        if is_classification:
            y = numpy.round(y, 0).astype(int)
        return DMD(x=x, y=y,
                   columns_meta={DMD.FEATURE_NAMES: ['f_' + str(k) for k in
                                                     range(x.shape[1])]})

    def get_model(self, is_classification):
        if is_classification:
            estimator = RandomForestClassifier
        else:
            estimator = RandomForestRegressor

        model = GeneralUtils.simple_imputation_pipeline(
            estimator(random_state=0, n_jobs=multiprocessing.cpu_count() - 1, n_estimators=15))

        return model

    def _test(self, is_classification, method):
        model = self.get_model(is_classification)

        train = self.get_data(is_classification)
        model.fit(train.values, train.target.ravel())

        test = self.get_data(is_classification, seed=1)

        if is_classification:
            uncertainty_model = UncertaintyModelClassifier(model=model,
                                                           uncertainty_method=method)
        else:
            uncertainty_model = UncertaintyModelRegressor(model=model,
                                                          uncertainty_method=method)

        uncertainty_model.fit(dmd_test=test, n_jobs=5, n_estimators=5)

        new_data = self.get_data(is_classification, seed=2)
        yp = uncertainty_model.predict(new_data)
        uncertainty = uncertainty_model.uncertainty(new_data)

        metric = sklearn.metrics.recall_score if is_classification else sklearn.metrics.r2_score
        kwargs = {'average': 'macro'} if is_classification else {}
        base_score = metric(y_true=new_data.target,
                            y_pred=yp, **kwargs)

        p50 = numpy.percentile(numpy.unique(uncertainty), 50)

        good = (uncertainty < p50).ravel()
        subset_good_score = metric(
            y_true=new_data.target[good], y_pred=yp[good], **kwargs)

        bad = (uncertainty > p50).ravel()
        subset_bad_score = metric(
            y_true=new_data.target[bad], y_pred=yp[bad], **kwargs)

        print(subset_bad_score, base_score, subset_good_score)
        self.assertGreater(subset_good_score, base_score)
        self.assertLess(subset_bad_score, base_score)

        pprint(uncertainty_model.uncertainty_analysis(dmd_train=train))


    def test_classification_confidence(self):
        self._test(is_classification=True, method='confidence')

    def test_classification_probability(self):
        self._test(is_classification=True, method='probability')

    def test_regression_mae(self):
        self._test(is_classification=False, method='mae')

    def test_regression_rmse(self):
        self._test(is_classification=False, method='rmse')

    def test_regression_quantile(self):
        self._test(is_classification=False, method='quantile')

        # subset_bad_score = []
        # subset_good_score = []
        # for k in range(10):
        #     bad = (uncertainty > (k+1)/11).ravel()
        #     subset_bad_score.append(sklearn.metrics.recall_score(
        #         y_true=new_data.target[bad], y_pred=yp[bad], average='macro'))
        #
        #     good = (uncertainty < (k + 1) / 11).ravel()
        #     subset_good_score.append(sklearn.metrics.recall_score(
        #         y_true=new_data.target[good], y_pred=yp[good], average='macro'))
        #
        # from matplotlib import pyplot as plt
        # plt.plot(range(10), subset_bad_score,'.-r',
        #          list(reversed(range(10))), subset_good_score,'.-g')
        # plt.show()



import logging

import numpy

from pytolemaic.analysis_logic.dataset_analysis.dataset_analysis import DatasetAnalysis, CovarianceShift
from pytolemaic.analysis_logic.dataset_analysis.dataset_analysis_report import DatasetAnalysisReport
from pytolemaic.analysis_logic.model_analysis.scoring.scoring import \
    Scoring
from pytolemaic.analysis_logic.model_analysis.scoring.scoring_report import ScoringFullReport
from pytolemaic.analysis_logic.model_analysis.sensitivity.sensitivity import \
    SensitivityAnalysis
from pytolemaic.analysis_logic.model_analysis.sensitivity.sensitivity_reports import SensitivityFullReport
from pytolemaic.analysis_logic.prediction_analysis.lime_report import LimeExplainer
from pytolemaic.analysis_logic.quality_report import TestSetQualityReport, TrainSetQualityReport, QualityReport, \
    ModelQualityReport
from pytolemaic.prediction_uncertainty.uncertainty_model import \
    UncertaintyModelClassifier, UncertaintyModelRegressor, UncertaintyModelBase
from pytolemaic.pytrust_report import PyTrustReport
from pytolemaic.utils.constants import CLASSIFICATION, REGRESSION, FeatureTypes
from pytolemaic.utils.dmd import DMD, ShuffleSplitter, StratifiedSplitter
from pytolemaic.utils.general import GeneralUtils
from pytolemaic.utils.metrics import Metrics, Metric


def cache(func):
    def cache_wrapper(self, *args, **kwargs):
        if not hasattr(self, '_cache'):
            self._cache = {}

        if func.__name__ not in self._cache:
            self._cache[func.__name__] = func(self, *args, **kwargs)

        return self._cache[func.__name__]

    return cache_wrapper


class PyTrust():
    numerical = FeatureTypes.numerical
    categorical = FeatureTypes.categorical

    def __init__(self, model,
                 xtrain=None, ytrain=None,
                 sample_meta_train: dict = None,

                 xtest=None, ytest=None,
                 sample_meta_test: dict = None,

                 columns_meta: dict = None,

                 feature_names: list = None,
                 feature_types: list = None,
                 categorical_encoding: dict = None,
                 metric: [str, Metric] = None,
                 splitter: str = 'shuffled',
                 target_labels: dict = None):
        """

        :param model: Model trained on training data provided

        :param xtrain: X training data. if DMD is provided, ytrain and any additional metadata is ignored.
        :param ytrain: Y training data.
        :param sample_meta_train: generic way to provide meta information on each sample in train data (e.g. sample weight) {key : [list of values]}.

        :param xtest: X test data. if DMD is provided, ytest and any additional metadata is ignored..
        :param ytest: Y test data. if DMD is provided,
        :param sample_meta_test: generic way to provide meta information on each sample in test data (e.g. sample weight) {key : [list of values]}.

        :param columns_meta: generic way to provide meta information on each feature (e.g. feature name) {key : [list of values]}.
        :param feature_names: feature name for each feature
        :param feature_types: feature type for each feature: NUMERICAL or CATEGORICAL
        :param categorical_encoding: For each column of categorical feature type, provide a dictionary of the structure
        {index: class name}. This information will allow providing more readable reports.

        :param metric: Target metric
        :param splitter: Splitter
        :param target_labels: categorical encoding for target variable in the format of {index: class name}.
        """
        self.model = model

        if isinstance(splitter, str):
            if splitter == 'shuffled':
                splitter = ShuffleSplitter
            elif splitter == 'stratified':
                splitter = StratifiedSplitter
            else:
                raise NotImplementedError("splitter='{}' is not supported".format(splitter))
        else:
            if not hasattr(splitter, 'split'):
                raise ValueError("splitter='{}' does not supported split() operation".format(splitter))
            else:
                raise NotImplementedError("splitter='{}' is not supported".format(splitter))

        if ytrain is not None:
            shape = getattr(ytrain, 'shape', (1, 1))
            if len(shape) == 2 and shape[1] > 1:
                raise NotImplementedError("Pytrust does not support multilabel (ytrain.shape[1]>1) analysis. "
                                          "In order to use Pytolemaic package, please wrap you model so model.predict(X) will return a single vector. ")
        if ytest is not None:
            shape = getattr(ytest, 'shape', (1, 1))
            if len(shape) == 2 and shape[1] > 1:
                raise NotImplementedError("Pytrust does not support multilabel (ytest.shape[1]>1) analysis. "
                                          "In order to use Pytolemaic package, please wrap you model so model.predict(X) will return a single vector. ")

        self.train = xtrain
        if self.train is not None and not isinstance(self.train, DMD):
            self.train = DMD(x=xtrain, y=ytrain,
                             samples_meta=sample_meta_train,
                             columns_meta=columns_meta,
                             feature_names=feature_names,
                             feature_types=feature_types,
                             categorical_encoding=categorical_encoding,
                             splitter=splitter,
                             target_labels=target_labels)

        self.test = xtest
        if self.test is not None and not isinstance(self.test, DMD):
            self.test = DMD(x=xtest, y=ytest,
                            samples_meta=sample_meta_test,
                            columns_meta=columns_meta,
                            feature_names=feature_names,
                            feature_types=feature_types,
                            categorical_encoding=categorical_encoding,
                            splitter=splitter,
                            target_labels=target_labels)

        if metric is None:
            if GeneralUtils.is_classification(model):
                metric = Metrics.recall
            else:
                metric = Metrics.mae

        self.metric = metric.name if isinstance(metric, Metric) else metric

        # todo
        self._validate_input()

        self.sensitivity = SensitivityAnalysis()

        self._uncertainty_models = {}
        self.covariance_shift = None
        self._cache = {}

    def _validate_input(self):
        if not hasattr(self.model, 'predict'):
            raise ValueError("Model must support predict() function")

    # region report classmethods
    @classmethod
    def create_sensitivity_report(cls, model, train: DMD, test: DMD, metric: str,
                                  sensitivity: SensitivityAnalysis = None, **kwargs) -> SensitivityFullReport:
        """
        Create sensitivity report
        Args:
            model - model to be analyzed based on test data
            train - train data (DMD). Train data is used if test is None
            test - test data (DMD). If None, use train data instead.
            sensitivity - SensitivityAnalysis instance. If None, instance is initiated internally.

        Returns:
            sensitivity report
        """
        sensitivity = sensitivity or SensitivityAnalysis()
        sensitivity.calculate_sensitivity(
            model=model, dmd_test=test, dmd_train=train, metric=metric, **kwargs)

        return sensitivity.sensitivity_report(**kwargs)

    @classmethod
    def create_scoring_report(cls, model, test: DMD, metric: str, y_pred=None, y_proba=None,
                              scoring: Scoring = None, **kwargs) -> ScoringFullReport:
        """
        Create scoring report
        Args:
            model - model to be analyzed based on test data
            train - train data (DMD). Train data is used only for separability test
            test - test data (DMD). .
            scoring - Scoring instance. If None, instance is initiated internally.

        Returns:
            scoring report
        """
        metrics = Metrics.supported_metrics()

        scoring = scoring or Scoring(metrics=metrics)

        score_values_report, confusion_matrix, scatter, classification_report = \
            scoring.score_value_report(model=model,
                                       dmd_test=test,
                                       labels=test.labels,
                                       y_pred=y_pred,
                                       y_proba=y_proba)

        return ScoringFullReport(target_metric=metric,
                                 metric_reports=score_values_report,
                                 confusion_matrix=confusion_matrix,
                                 scatter=scatter,
                                 classification_report=classification_report)

    @classmethod
    def create_quality_report(cls, scoring_report: ScoringFullReport,
                              sensitivity_report: SensitivityFullReport,
                              dataset_analysis_report: DatasetAnalysisReport) -> QualityReport:
        """
        Create quality report by analyzing scoring_report & sensitivity_report

        Returns:
            quality report
        """
        test_set_report = TestSetQualityReport(scoring_report=scoring_report,
                                               dataset_analysis_report=dataset_analysis_report)

        train_set_report = TrainSetQualityReport(vulnerability_report=sensitivity_report.vulnerability_report)
        model_quality_report = ModelQualityReport(vulnerability_report=sensitivity_report.vulnerability_report,
                                                  scoring_report=scoring_report)

        return QualityReport(train_quality_report=train_set_report, test_quality_report=test_set_report,
                             model_quality_report=model_quality_report)

    @classmethod
    def create_dataset_analysis_report(cls, train: DMD, is_classification, test:DMD=None, dataset_analysis: DatasetAnalysis=None, **kwargs) -> DatasetAnalysisReport:
        """
        Create dataset analysis report by analyzing train data

        Args:
            train - train data
            test - test data, if available
            is_classification - whether the target is categorical.

        Returns:
            dataset analysis report
        """
        dataset_analysis = dataset_analysis or DatasetAnalysis(problem_Type=CLASSIFICATION if is_classification else REGRESSION)
        report = dataset_analysis.dataset_analysis_report(train=train, test=test)
        return report

    @classmethod
    def create_pytrust_report(cls, pytrust) -> PyTrustReport:
        """
             Creates final report encompassing all other reports for entire pytrust.
             Returns:
                 pytrust report
        """
        return PyTrustReport(pytrust=pytrust)

    @classmethod
    def create_pytrust_for_separability_test(cls, train: DMD, test: DMD, covariance_shift=None, **pytrust_kwargs):
        """
        Create scoring report
        Args:
            train - train data (DMD).
            test - test data (DMD).

        Returns:
            separability pytrust
        """

        covariance_shift = covariance_shift or CovarianceShift()
        covariance_shift.calc_covariance_shift(dmd_train=train, dmd_test=test)

        separability_pytrust = PyTrust(model=covariance_shift.classifier,
                                       xtrain=covariance_shift.train,
                                       xtest=covariance_shift.test,
                                       **pytrust_kwargs)
        return separability_pytrust

    # endregion

    # region reports
    def _create_sensitivity_report(self) -> SensitivityFullReport:
        return self.create_sensitivity_report(model=self.model, train=self.train, test=self.test, metric=self.metric,
                                              sensitivity=self.sensitivity)

    def _create_scoring_report(self) -> ScoringFullReport:
        metrics = Metrics.supported_metrics()

        self.scoring = Scoring(metrics=metrics)

        return self.create_scoring_report(model=self.model, train=self.train, test=self.test, metric=self.metric,
                                          y_pred=self.y_pred_test, y_proba=self.y_proba_test, scoring=self.scoring)

    def _create_quality_report(self) -> QualityReport:
        return self.create_quality_report(scoring_report=self.scoring_report,
                                          sensitivity_report=self.sensitivity_report,
                                          dataset_analysis_report=self.dataset_analysis_report)

    def _create_dataset_analysis_report(self, **kwargs) -> DatasetAnalysisReport:
        self.dataset_analysis = DatasetAnalysis(problem_Type=CLASSIFICATION if self.is_classification else REGRESSION)
        return self.create_dataset_analysis_report(train=self.train, test=self.test, is_classification=self.is_classification,
                                                   dataset_analysis=self.dataset_analysis)

    def _create_pytrust_report(self):
        return self.create_pytrust_report(pytrust=self)

    def _create_pytrust_for_separability_test(self):
        self.covariance_shift = CovarianceShift()
        return self.create_pytrust_for_separability_test(
            train=self.train, test=self.test, metric=self.metric, covariance_shift=self.covariance_shift)

    # endregion

    # region on predict

    def create_uncertainty_model(self, method='default') -> UncertaintyModelBase:
        if method not in self._uncertainty_models:

            if self.is_classification:
                uncertainty_model = UncertaintyModelClassifier(
                    model=self.model, uncertainty_method=method)
            else:
                uncertainty_model = UncertaintyModelRegressor(
                    model=self.model, uncertainty_method=method)

            uncertainty_model.fit(dmd_test=self.test)
            self._uncertainty_models[method] = uncertainty_model

        return self._uncertainty_models[method]

    @cache
    def create_lime_explainer(self, **kwargs):
        lime_explainer = LimeExplainer(n_features_to_plot=20, **kwargs)
        lime_explainer.fit(self.train, model=self.model)
        return lime_explainer

    # endregion

    # region properties

    @property
    @cache
    def is_classification(self):
        return GeneralUtils.is_classification(self.model)

    @property
    @cache
    def model_support_dmd(self):
        return GeneralUtils.dmd_supported(self.model, self.test)

    @property
    @cache
    def y_pred_test(self):

        test = self.test if self.model_support_dmd else self.test.values
        if self.y_proba_test is not None:  # save some time
            y_pred_test = numpy.argmax(self.y_proba_test, axis=1)
        else:
            y_pred_test = self.model.predict(test)
        return y_pred_test

    @property
    @cache
    def y_proba_test(self):
        if not self.is_classification:
            return None

        test = self.test if self.model_support_dmd else self.test.values
        y_proba_test = self.model.predict_proba(test)
        return y_proba_test

    @property
    @cache
    def sensitivity_report(self)-> SensitivityFullReport:
        return self._create_sensitivity_report()

    @property
    @cache
    def scoring_report(self)-> ScoringFullReport:
        return self._create_scoring_report()

    @property
    @cache
    def quality_report(self)-> QualityReport:
        return self._create_quality_report()

    @property
    @cache
    def dataset_analysis_report(self)->DatasetAnalysisReport:
        return self._create_dataset_analysis_report()

    @property
    @cache
    def report(self) -> PyTrustReport:
        return self._create_pytrust_report()

    @property
    @cache
    def insights(self) -> list:
        return self.report.insights()

    @property
    @cache
    def pytrust_for_separability_test(self):
        return self._create_pytrust_for_separability_test()

    @classmethod
    def print_initialization_example(cls):
        example = \
            """
        # example for initializing PyTrust object. In this example, All fields other than model, xtrain, ytrain are optional.
        PyTrust(model=model,
                xtrain=X_train,
                ytrain=y_train,
                xtest=X_test,
                ytest=Y_test,
                feature_names=['Feature_#{}'.format(i) for i in range(X_train.shape[1])],
                feature_types=[PyTrust.categorical for i in range(X_train.shape[1])],
                target_labels={0:'dog', 1:'cat'}, # for classification
                categorical_encoding=[{0:'blue', 1:'red', 2:'yellow'} for i in range(X_train.shape[1])]                                
                )
                
        # See usage example below:\n
        """
        logging.info(example)
        example += cls.print_usage_example()
        return example

    @classmethod
    def print_usage_example(cls, report_types=('report',), plot=True, insights=True, to_dict=True,
                            to_dict_meaning=True):
        def plot_usage(prefix):
            return "# Plotting relevant graphs:\n" \
                   "{}.plot()\n" \
                   "plt.show()".format(prefix)

        def insights_usage(prefix):
            return "# Printing useful insights & hints. Short and to the point.\n" \
                   "pprint({}.insights())".format(prefix)

        def to_dict_usage(prefix):
            return "# Prints all information gathered, either full (default) or in readable form (printable=True)\n" \
                   "pprint({}.to_dict())\n" \
                   "or\n" \
                   "pprint({}.to_dict(printable=True))".format(prefix, prefix)

        def to_dict_meaning_usage(prefix):
            return "# Prints explanation for every field within the dict returned by to_dict()\n" \
                   "pprint({}.to_dict_meaning())".format(prefix)

        def add_components_msgs(prefix, plot=True, insights=True, to_dict=True, to_dict_meaning=True):
            msgs = []
            if plot:
                msgs.append(plot_usage(prefix=prefix))

            if insights:
                msgs.append(insights_usage(prefix=prefix))

            if to_dict:
                msgs.append(to_dict_usage(prefix=prefix))

            if to_dict_meaning:
                msgs.append(to_dict_meaning_usage(prefix=prefix))

            return "\n\n".join(msgs)

        example = []
        for report in report_types:
            if report not in ['report']:
                raise NotImplementedError(report)

            example.append('\n*****  pytrust.{}  ******\n'.format(report))

            if report == 'report':
                msg = add_components_msgs(prefix="pytrust.report",
                                          plot=plot, insights=insights,
                                          to_dict=to_dict, to_dict_meaning=to_dict_meaning)
                if msg:
                    example.append(msg)

            example.append('\n**************\n')

        example = "\n".join(example)
        logging.info(example)
        return example


    # endregion



import logging

import numpy
import scipy.stats

from pytolemaic.utils.constants import CLASSIFICATION
from pytolemaic.utils.dmd import DMD
from pytolemaic.analysis_logic.dataset_analysis.covriance_shift import CovarianceShift
from pytolemaic.analysis_logic.dataset_analysis.dataset_analysis_report import \
    DatasetAnalysisReport, MissingValuesReport



class DatasetAnalysis():
    def __init__(self, problem_Type, class_count_threshold=10, outliers_n_sigma=(3, 5),
                 nan_threshold_per_col=(0.1, 0.5, 0.9), nan_threshold_per_sample=(0.1, 0.5, 0.9)):
        self._problem_type = problem_Type
        self._class_count_threshold = class_count_threshold
        self._outliers_n_sigma = outliers_n_sigma
        self._nan_threshold_per_feature = nan_threshold_per_col
        self._nan_threshold_per_sample = nan_threshold_per_sample
        self._covariance_shift = None

    def count_unique_classes(self, dataset: DMD) -> dict:

        if dataset.categorical_features is None:
            logging.warning("Unable to analyze categorical features if feature types information is unavailable.")
            return {}

        if len(dataset.categorical_features) == 0:
            return {}

        out = {}
        x = numpy.zeros((dataset.n_samples, dataset.n_features + 1))
        x[:, :-1] = dataset.values
        x[:, -1] = dataset.target.ravel()

        feature_names = dataset.feature_names + ['target']

        nan_mask = numpy.zeros((dataset.n_samples, dataset.n_features + 1)).astype(bool)
        nan_mask[:, :-1] = dataset.nan_mask

        target_is_categorical = self._problem_type == CLASSIFICATION
        target_index = [dataset.n_features] if target_is_categorical else []

        categorical_features = dataset.categorical_features.tolist() + target_index

        for i in categorical_features:
            vec = x[~nan_mask[:, i], i]
            if len(vec) == 0:
                continue
            try:
                classes_, counts = numpy.unique(vec, return_counts=True)
            except:
                classes_ = set(vec)
                counts = [sum(vec == class_) for class_ in classes_]

            if i in dataset.categorical_encoding_by_icols:
                encoding = dataset.categorical_encoding_by_icols[i]
            elif i == target_index and dataset.target_encoding is not None:
                encoding = dataset.target_encoding
            else:
                encoding = None

            if encoding is not None:
                classes_ = [encoding[i] for i in classes_]

            if min(counts) <= self._class_count_threshold:
                out[feature_names[i]] = {class_: count for class_, count in zip(classes_, counts) if
                                         count <= self._class_count_threshold}

        return out

    def count_outliers(self, dataset: DMD) -> dict:
        if dataset.categorical_features is None:
            logging.warning(
                "Unable to analyze numerical features for outlier if feature types information is unavailable.")
            return {}

        if len(dataset.categorical_features) == dataset.n_features:
            return {}

        out = {}
        x = numpy.zeros((dataset.n_samples, dataset.n_features + 1))
        x[:, :-1] = dataset.values
        x[:, -1] = dataset.target.ravel()
        feature_names = dataset.feature_names + ['target']

        nan_mask = numpy.zeros((dataset.n_samples, dataset.n_features + 1)).astype(bool)
        nan_mask[:, :-1] = dataset.nan_mask

        target_is_categorical = self._problem_type == CLASSIFICATION
        target_index = {dataset.n_features} if target_is_categorical else set()
        numerical_features = set(numpy.arange(dataset.n_features + 1)) - set(
            dataset.categorical_features) - target_index

        expected_outliers_per_sigma = {sigma: int(0.75 + (1 - scipy.stats.norm.cdf(sigma)) * dataset.n_samples)
                                       for sigma in self._outliers_n_sigma}

        for i in numerical_features:
            vec = x[~nan_mask[:, i], i]
            if len(vec) == 0:
                continue

            for sigma in self._outliers_n_sigma:
                mean, std = self._calc_mean_and_std(sigma, vec)
                mn, mx = numpy.min(vec), numpy.max(vec)

                expected_outliers = expected_outliers_per_sigma[sigma]
                n_outliers = numpy.sum(vec > mean + sigma * std) + numpy.sum(vec < mean - sigma * std)

                if n_outliers > 2 * expected_outliers:
                    if feature_names[i] not in out:
                        out[feature_names[i]] = {}
                    out[feature_names[i]]['{}-sigma'.format(sigma)] = dict(n_outliers=n_outliers,
                                                                           n_sigma=sigma,
                                                                           expected_outliers=expected_outliers,
                                                                           mean=mean,
                                                                           std=std,
                                                                           min=mn,
                                                                           max=mx)
                    # n_sigma = '{}-sigma'.format(sigma)
                    # if n_sigma not in out:
                    #     out[n_sigma] = {}
                    # out[n_sigma][feature_names[i]] = n_outliers

        return out

    def _calc_mean_and_std(self, sigma, vec):
        # Outliers will strongly affect the mean and variance by which they are defined.
        # Thus, we throw away outliers according to sigma+2 before calculating the mean and std.
        # We repeat the process iteratively until converged.

        prev_indices = []
        indices = numpy.arange(len(vec))
        mean, std = 0, 0
        while len(prev_indices) != len(indices):
            vec_for_stats = vec[indices]
            std = numpy.std(vec_for_stats)
            mean = numpy.mean(vec_for_stats)
            outliers = ((vec_for_stats > mean + (2 + sigma) * std) + (
                    vec_for_stats < mean - (2 + sigma) * std)).astype(bool)
            prev_indices = indices
            indices = indices[~outliers]
        return mean, std

    def count_missing_values(self, dataset: DMD):
        nan_mask = dataset.nan_mask

        # features
        nan_cols = {}
        for nan_threshold in self._nan_threshold_per_feature:
            nan_ratios = numpy.sum(nan_mask, axis=0) / dataset.n_samples
            is_off_limit = nan_ratios >= nan_threshold
            nan_cols[nan_threshold] = dict(
                zip(numpy.array(dataset.feature_names)[is_off_limit], nan_ratios[is_off_limit]))

        # samples
        nan_rows = {}
        for nan_threshold in self._nan_threshold_per_sample:
            nan_ratios = numpy.sum(nan_mask, axis=1) / dataset.n_features
            is_off_limit = nan_ratios >= nan_threshold
            nan_rows[nan_threshold] = dict(zip(numpy.arange(dataset.n_samples)[is_off_limit], nan_ratios[is_off_limit]))

        return nan_cols, nan_rows

    @property
    def covariance_shift(self):
        return self._covariance_shift

    def dataset_analysis_report(self, train: DMD, test:DMD=None) -> DatasetAnalysisReport:
        nan_counts_features, nan_counts_samples = self.count_missing_values(dataset=train)

        self.count_unique_classes(dataset=train)
        self.count_outliers(dataset=train)
        if test is None:
            covariance_report = None
        else:
            self._covariance_shift = CovarianceShift()
            self._covariance_shift.calc_covariance_shift(dmd_train=train, dmd_test=test)
            covariance_report = self._covariance_shift.covariance_shift_report()

        return DatasetAnalysisReport(
            class_counts=self.count_unique_classes(dataset=train),
            outliers_count=self.count_outliers(dataset=train),
            missing_values_report=MissingValuesReport(nan_counts_features=nan_counts_features,
                                                      nan_counts_samples=nan_counts_samples),
            covariance_shift_report=covariance_report
        )



import copy
import logging

import numpy
import numpy as np
import pandas
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder

from pytolemaic.utils.constants import FeatureTypes


class ShuffleSplitter():
    @classmethod
    def split(cls, dmdy, ratio=0.1, random_state=0):
        train, test = train_test_split(numpy.arange(dmdy.n_samples),
                                       test_size=ratio, random_state=random_state,
                                       shuffle=True, stratify=None)

        return train, test


class StratifiedSplitter():
    @classmethod
    def split(cls, dmdy, ratio=0.1, random_state=0):
        train, test = train_test_split(numpy.arange(dmdy.n_samples),
                                       test_size=ratio, random_state=random_state,
                                       shuffle=True, stratify=dmdy.target)

        if len(test) < ratio * 0.95 * dmdy.n_samples:
            logging.warning("Issue encountered with sklearn's stratified split! Reveting to shuffle split")
            train, test = ShuffleSplitter.split(dmdy=dmdy, ratio=ratio, random_state=random_state)

        return train, test



class DMD():
    FEATURE_NAMES = '__FEATURE_NAMES__'
    FEATURE_TYPES = '__FEATURE_TYPES__'
    INDEX = '__INDEX__'
    CATEGOICAL_ENCODING = '__CATEGOICAL_ENCODING__'

    # SAMPLE_WEIGHTS = '__SAMPLE_WEIGHTS__'

    def __init__(self, x, y=None, columns_meta: dict = None, samples_meta: dict = None,
                 splitter=ShuffleSplitter, target_labels: dict = None, categorical_encoding: dict = None,
                 feature_names: list = None, feature_types: list = None):

        self._x = pandas.DataFrame(x)
        if y is not None:
            self._y = pandas.DataFrame(y)
        else:
            self._y = None

        self._columns_meta = self._create_columns_meta(columns_meta, self._x, feature_names=feature_names, feature_types=feature_types)
        self._x.columns = self._columns_meta[self.FEATURE_NAMES]

        self._samples_meta = self._create_samples_meta(samples_meta, self._x)
        self._splitter = splitter

        # meta data
        if target_labels is not None:
            if self._y.values.max() >= len(target_labels):
                raise ValueError("Labels should be given to all classes")

        if isinstance(target_labels, list):
            target_labels = {i: cls_ for i, cls_ in enumerate(target_labels)}

        self._target_encoding = target_labels

        self._categorical_encoding_by_name = categorical_encoding or {}
        self._categorical_encoding_by_ind = self._validate_categorical_encoding()

    def _validate_categorical_encoding(self):
        categorical_encoding = self.categorical_encoding_by_feature_name

        if len(categorical_encoding) == 0:
            return {}

        categorical_features = self.categorical_features
        if categorical_features is None or len(categorical_features) == 0:
            raise ValueError(
                "When setting categorical_encoding you must also specify which feature is categorical through columns_meta")

        feature_names = self.feature_names
        values = self.values.astype(float, copy=False)
        for icol in categorical_features:
            if feature_names[icol] not in categorical_encoding:
                raise ValueError("No categorical names are set for feature #{}:{}".format(icol, feature_names[icol]))

            vec = values[:, icol]
            vec = vec[numpy.isfinite(vec)]
            delta = set(vec) - set(categorical_encoding[feature_names[icol]].keys())
            if delta:
                raise ValueError("No categorical name is set for categories {} (feature #{}:{})".format(delta, icol,
                                                                                                        feature_names[
                                                                                                            icol]))

        return {icol: self._categorical_encoding_by_name[feature_names[icol]] for icol in categorical_features}


    def __deepcopy__(self, memodict={}):
        return type(self)(x=copy.deepcopy(self._x),
                          y=copy.deepcopy(self._y),
                          columns_meta=copy.deepcopy(self._columns_meta),
                          samples_meta=copy.deepcopy(self._samples_meta),
                          splitter=copy.deepcopy(self._splitter),
                          )

    @classmethod
    def _create_columns_meta(cls, columns_meta, df, feature_names=None, feature_types=None)->pandas.DataFrame:
        # columns_meta is either pd.DataFrame or dict
        if columns_meta is None:
            columns_meta = {}

        if cls.FEATURE_NAMES not in columns_meta:
            if feature_names:
                columns_meta[cls.FEATURE_NAMES] = feature_names
            else:
                columns_meta[cls.FEATURE_NAMES] = df.columns

        if cls.FEATURE_TYPES not in columns_meta and feature_types is not None:
            columns_meta[cls.FEATURE_TYPES] = feature_types

        columns_meta = pandas.DataFrame(columns_meta)

        if df.shape[1] != columns_meta.shape[0]:
            raise ValueError("Given data has {} features but columns metadata was given for {} features"
                             .format(df.shape[1], columns_meta.shape[0]))
        return columns_meta

    @classmethod
    def _create_samples_meta(cls, samples_meta, df):
        if samples_meta is None:
            samples_meta = pandas.DataFrame(
                {DMD.INDEX: np.arange(df.shape[0])})
        else:
            samples_meta = pandas.DataFrame(samples_meta)

        if DMD.INDEX not in samples_meta:
            samples_meta[DMD.INDEX] = np.arange(df.shape[0])
        return samples_meta

    def split_by_indices(self, indices):
        return DMD(x=copy.deepcopy(self._x.iloc[indices, :]),
                   y=copy.deepcopy(self._y.iloc[indices, :]),
                   columns_meta=copy.deepcopy(self._columns_meta),
                   samples_meta=copy.deepcopy(
                       self._samples_meta.iloc[indices, :]),
                   splitter=self.splitter,
                   target_labels=self.target_encoding,
                   categorical_encoding=self.categorical_encoding_by_feature_name)

    def split(self, ratio, return_indices=False):

        dmd_y = DMD(x=self._y.values if self._y is not None else np.arange(
            self.n_samples), samples_meta=self._samples_meta)
        left, right = self.splitter.split(dmd_y, ratio=ratio)
        left = list(sorted(left))
        right = list(sorted(right))
        if return_indices:
            return left, right
        else:
            return self.split_by_indices(left), self.split_by_indices(right),

    def append(self, other, axis=0):
        if axis == 0:
            if other.feature_names != self.feature_names:
                raise ValueError("Cannot concat DMDs due to feature names difference")
            # noinspection PyTypeChecker
            if any(self._samples_meta.columns != other._samples_meta.columns):
                raise ValueError("Cannot concat DMDs due to samples meta difference")

            self._x = pandas.concat([self._x, other._x], axis=axis, ignore_index=True, copy=True)
            self._y = pandas.concat([self._y, other._y], axis=axis, ignore_index=True, copy=True)
            self._samples_meta = pandas.concat([self._samples_meta, other._samples_meta], axis=axis, ignore_index=True, copy=True)
            self._samples_meta[self.INDEX] = np.arange(self.n_samples)

        else:
            raise NotImplementedError("not implemented yet")

    def set_target(self, new_y):
        if new_y is None:
            self._y = None
        elif len(new_y) != self.n_samples:
            raise ValueError("Mismatch in number of samples")

        self._y = pandas.DataFrame(new_y)

    @classmethod
    def concat(cls, dmds, axis=0):
        dmd = None
        for dmd_item in dmds:
            if dmd is None:
                dmd = copy.deepcopy(dmd_item)
            else:
                dmd.append(dmd_item, axis=axis)

        return dmd

    @property
    def feature_names(self):
        return list(self._columns_meta[DMD.FEATURE_NAMES])

    @property
    def shape(self):
        return self._x.shape

    @property
    def n_samples(self):
        return self.shape[0]

    @property
    def n_features(self):
        return self.shape[1]

    @property
    def index(self):
        return self._samples_meta[DMD.INDEX]

    @property
    def values(self):
        return self._x.values

    @property
    def target(self):
        if self._y is None:
            return None
        else:
            return self._y.values

    @property
    def splitter(self):
        return self._splitter

    @property
    def target_encoding(self) -> dict:
        return self._target_encoding  # {index: cls_}

    @property
    def labels(self) -> list:
        """
        :return: class names ordered by encoding.
        """
        if self.target_encoding is None:
            return None

        encoding = self._target_encoding  # {index: cls_}
        inv_encoding = {cls_: index for index, cls_ in encoding.items()}
        labels = sorted(inv_encoding.keys(), key=lambda cls_: inv_encoding[cls_])
        return labels

    @property
    def categorical_features(self):
        if self.FEATURE_TYPES not in self._columns_meta.columns:
            return None
        else:
            return numpy.arange(self.n_features)[
                self.feature_types == FeatureTypes.categorical]

    @property
    def numerical_features(self):
        if self.FEATURE_TYPES not in self._columns_meta.columns:
            return None
        else:
            return numpy.arange(self.n_features)[
                self.feature_types == FeatureTypes.numerical]

    @property
    def feature_types(self):
        if self.FEATURE_TYPES not in self._columns_meta.columns:
            return None
        else:
            return self._columns_meta[self.FEATURE_TYPES].values.ravel()

    @property
    def categorical_encoding_by_feature_name(self):
        return self._categorical_encoding_by_name

    @property
    def categorical_encoding_by_icols(self):
        return self._categorical_encoding_by_ind

    @property
    def nan_mask(self):
        return self._x.isnull().values

    @classmethod
    def from_df(cls, df_train: pandas.DataFrame, target_name: str, is_classification, feature_types: list = None,
                df_test: pandas.DataFrame = None,
                categorical_encoding=False, split_ratio: float = None, nan_list: list = (), splitter=ShuffleSplitter):
        feature_names = list(df_train.columns)
        feature_names.remove(target_name)

        categorical_encoding_dict = None
        target_labels = None
        if categorical_encoding:
            if feature_types is None:
                raise ValueError("categorical_encoding requires feature types")
            logging.info("Categorical Encoding")
            # fit
            jibrish_value = '!@#$%^&*()'
            df_train = copy.deepcopy(df_train)
            for k in nan_list:
                df_train.replace(k, numpy.nan, inplace=True)

            nan_mask = df_train.isnull()
            # df_train[nan_mask] = jibrish_value
            df_train.replace(numpy.nan, jibrish_value, inplace=True)

            xencoders = []
            categorical_encoding_dict = {}
            for i, feature_name in enumerate(feature_names):
                logging.info("Fit encoding for feature {}".format(feature_name))

                if feature_types[i] == FeatureTypes.categorical:
                    le = LabelEncoder().fit(df_train[feature_name].astype(str))
                    xencoders.append(le)
                    categorical_encoding_dict[feature_name] = {i: cls_ for i, cls_ in enumerate(le.classes_)}
                else:
                    xencoders.append(None)

            if is_classification:
                yencoder = LabelEncoder().fit(df_train[target_name].astype(str))
                target_labels = {i: cls_ for i, cls_ in enumerate(yencoder.classes_)}
            else:
                yencoder = None
                target_labels = None

            # transform train
            for i, feature_name in enumerate(feature_names):
                if feature_types[i] == FeatureTypes.categorical:
                    logging.info("Encode train: feature {}".format(feature_name))
                    df_train[feature_name] = xencoders[i].transform(df_train[feature_name].astype(str))

            if yencoder:
                df_train[target_name] = yencoder.transform(df_train[target_name].astype(str))

            df_train[nan_mask] = numpy.nan

            if df_test is not None:
                df_test = copy.deepcopy(df_test)
                for k in nan_list:
                    df_test.replace(k, numpy.nan, inplace=True)

                nan_mask = df_test.isnull()
                df_test.replace(numpy.nan, jibrish_value, inplace=True)

                for i, feature_name in enumerate(feature_names):
                    if feature_types[i] == FeatureTypes.categorical:
                        logging.info("Encode test: feature {}".format(feature_name))

                        try:
                            df_test[feature_name] = xencoders[i].transform(df_test[feature_name].astype(str))
                        except ValueError as e:
                            logging.warning(e)
                            classes = xencoders[i].classes_
                            mask = numpy.array([x in classes for x in df_test[feature_name].astype(str)])

                            new_values = numpy.zeros(mask.shape)
                            new_values[mask] = xencoders[i].transform(df_test[feature_name].astype(str)[mask])
                            new_values[~mask] = -1
                            df_test[feature_name] = new_values

                if yencoder:
                    df_test[target_name] = yencoder.transform(df_test[target_name].astype(str))

                df_test[nan_mask] = numpy.nan

        logging.info("Creating DMDs")

        dmd_train = DMD(x=df_train[feature_names], y=df_train[target_name], splitter=splitter,
                        target_labels=target_labels, categorical_encoding=categorical_encoding_dict,
                        feature_names=feature_names, feature_types=feature_types)
        if df_test is None:
            if split_ratio is None:
                dmd_test = None
            else:
                dmd_train, dmd_test = dmd_train.split(ratio=split_ratio)
        else:

            dmd_test = DMD(x=df_test[feature_names], y=df_test[target_name], splitter=splitter,
                           target_labels=target_labels, categorical_encoding=categorical_encoding_dict,
                           feature_names=feature_names, feature_types=feature_types)

        return dmd_train, dmd_test

    def to_df(self, copy=False):
        return self._x.copy(deep=copy), None if self._y is None else self._y.copy(deep=copy)


import numpy
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

from pytolemaic.utils.general import GeneralUtils


class LinearBase():
    def __init__(self, model, n_samples=500, n_features=5):
        self.n_samples = n_samples
        self.n_features = n_features

        self._xtrain, self._ytrain = None, None
        self.model = model
        self.random_state = numpy.random.RandomState(0)
        self.y_type = float


    def _function(self, x):
        y = 0 * x[:, 0] + 3 * x[:, 1] + numpy.sum(x[:, 2:], axis=1)
        if self.y_type == float:
            return y
        else:
            return y.astype(int)

    def column_names(self):
        # 1st has no importance, while 3rd has double importance

        return ['zero importance', 'triple importance'] + ['regular importance #%d' % (k + 1) for k in
                                                                    range(self.n_features - 2)]

    @property
    def training_data(self):
        if self._xtrain is None:
            self._xtrain, self._ytrain = self.get_samples()

        return self._xtrain, self._ytrain

    def get_samples(self):
        # Let's define dataset and label

        x = self.random_state.rand(self.n_samples, self.n_features)

        y = self._function(x)

        x = GeneralUtils.add_nans(x)

        return x, y

    def get_model(self):
        x, y = self.training_data
        self.model.fit(x, y)
        return self.model

class LinearClassificationDataset(LinearBase):
    def __init__(self, **kwargs):
        super(LinearClassificationDataset, self).__init__(
            model=GeneralUtils.simple_imputation_pipeline(
                RandomForestClassifier(random_state=0, n_estimators=3, n_jobs=1)),
            **kwargs)
        self.y_type = int

class LinearRegressionDataset(LinearBase):
    def __init__(self, **kwargs):
        super(LinearRegressionDataset, self).__init__(
            model=GeneralUtils.simple_imputation_pipeline(
                RandomForestRegressor(random_state=0, n_estimators=3, n_jobs=1)),
            **kwargs)
        self.y_type = float






from enum import Enum


class EnumBase(Enum):

    @classmethod
    def keys(cls, members=False):
        if members:
            return [item for item in cls]
        else:
            return [item.name for item in cls]

    @classmethod
    def values(cls):
        return [item.value for item in cls]

    @classmethod
    def asdict(cls):
        return {item.name: item.value for item in cls}



from pytolemaic.utils.general import GeneralUtils


class Report():
    def to_dict(self, printable=False):
        raise NotImplementedError

    @classmethod
    def to_dict_meaning(cls):
        raise NotImplementedError

    def plot(self):
        raise NotImplementedError

    def insights(self):
        raise NotImplementedError

    def _add_cls_name_prefix(self, l):
        prefix = str(type(self).__name__)
        return [prefix + ': ' + item for item in l]

    @classmethod
    def _printable_dict(cls, out, printable=False):
        if printable:
            return GeneralUtils.make_dict_printable(out)
        else:
            return out



class FeatureTypes():
    numerical = 'numerical'
    categorical = 'categorical'


REGRESSION = 'regression'
CLASSIFICATION = 'classification'

NUMERICAL = FeatureTypes.numerical
CATEGORICAL = FeatureTypes.categorical
